//
// o c t o p u s 3 . c
//
// "octopus", an educational 8bit virtual machine, type 3
//
// Jan. 13, 2002 by Wataru Nishida (http://www.skyfree.org)
//

#include <stdio.h>		// fprintf()
#include <stdlib.h>		// atoi(), exit()
#include <sys/io.h>		// ioperm(), inb(), outb()
#include <signal.h>		// signal(), raise(), SIG*
#include <unistd.h>		// usleep(), alarm()
#include <fcntl.h>		// open(), O_RDONLY
#include <termios.h>		// termios{}, tcgetattr(), tcsetattr()
#include <unistd.h>		// read(), close()
#include <sys/ioctl.h>		// ioctl()
#include <linux/kd.h>		// KDSKBMODE, K_XLATE, K_MEDIUMRAW
#include <string.h>		// strcmp()

// Message-related macros and constants

#define message(fmt, args...)	fprintf(logfile, fmt, ## args)
#define imessage(fmt, args...)	if (output_inst) fprintf(logfile, fmt, ## args)

int fmt_text_bytes = 32;	// How many text bytes to be displayed
int fmt_port_bytes = 16;	// How many port bytes to be displayed
int fmt_base       = 16;	// Base value for number formatting (2|10|16)
int fmt_zero_pad   = 1;		// Pad with zeroes?

// "octopus" status code

#define OCT_BREAK	-1	// Software break for debugging
#define OCT_HALTED	-2	// System is halted
#define OCT_ILLCOD	-3	// Illegal operation code
#define OCT_ILLMOD	-4	// Illegal operation mode
#define OCT_OVRRUN	-5	// Over-run error
#define OCT_KEYERR	-6	// Keyboard error

// "octopus" operand mode

#define OPER_IMM	0	// Immediate value
#define OPER_REG	1	// Register
#define OPER_IMM_REG	2	// Immediate / Register
#define OPER_REG_REG	3	// Register  / Register
#define OPER_REG_MEM	4	// Register  / Memory
#define OPER_MEM_REG	5	// Memory    / Register
#define OPER_IND_REG	6	// Indirect  / Register
#define OPER_REG_IND	7	// Register  / Indirect

// "octopus" operation code

#define OPEC_NOP	0	// nop
#define OPEC_MOV	1	// mov
#define OPEC_IN		2	// in
#define OPEC_OUT	3	// out
#define OPEC_MOVI	4	// movi
#define OPEC_ADDI	5	// addi
#define OPEC_SUBI	6	// subi
#define OPEC_MULI	7	// muli
#define OPEC_DIVI	8	// divi
#define OPEC_ANDI	9	// andi
#define OPEC_ORI	10	// ori
#define OPEC_TESTI	11	// testi
#define OPEC_JMP	12	// jmp
#define OPEC_JZ		13	// jz
#define OPEC_JNZ	14	// jnz
#define OPEC_JSR	15	// jsr
#define OPEC_PUSH	16	// push
#define OPEC_POP	17	// pop
#define OPEC_CALL	18	// call
#define OPEC_RET	19	// ret
#define OPEC_BREAK	30	// break
#define OPEC_HLT	31	// hlt

// "octopus" virtual port definition

#define VPORT_SOUND	0	// Sound control
#define VPORT_FREQ_HIGH	1	// Timer count (higher byte)
#define VPORT_FREQ_LOW	2	// Timer count (lower byte)
#define VPORT_KEY_LED	3	// Keyboard LED control
#define VPORT_KEY_AK	4	// Keyboard "ASDFGHJK" status bit

// The heart of "octopus"

typedef unsigned char byte;	// "byte" type definition

byte reg[ 16 ];			// General purpose registers
byte text[ 256 ];		// Code and data area
byte port[ 256 ];		// I/O port area
byte flag;			// Flag bits

int enable_key = 0;		// Enable keyboard I/O function
int output_inst = 1;		// Output switch for instructions
int max_exec_time = 2;		// Limit execution time (sec)
int max_exec_steps = 200;	// Limit execution steps
int steps = 0;			// How many steps were executed?
int nop_wait = 200000;		// Wait for NOP (200 msec)
int fdcon;			// File descriptor of /dev/console
FILE* logfile;			// Log file

#define REG_PC	7		// Program counter
#define pc	reg[ REG_PC ]	// Program counter macro
#define REG_RET	6		// Return address holder
#define rad	reg[ REG_RET ]	// Return address macro
#define REG_SP	5		// Stack pointer
#define sp	reg[ REG_SP ]	// Stack pointer macro
#define ZFLAG	1		// Zero flag

// "octopus" instruction table definition

typedef struct {
  const char name[ 10 ];	// Instructon name
  int opecode;			// Opecode number
  int (*func)(int);		// Function pointer
 } instruction;

int oct_nop(int mode);
int oct_mov(int mode);
int oct_in(int mode);
int oct_out(int mode);
int oct_movi(int mode);
int oct_addi(int mode);
int oct_subi(int mode);
int oct_muli(int mode);
int oct_divi(int mode);
int oct_andi(int mode);
int oct_ori(int mode);
int oct_testi(int mode);
int oct_jmp(int mode);
int oct_jz(int mode);
int oct_jnz(int mode);
int oct_jsr(int mode);
int oct_push(int mode);
int oct_pop(int mode);
int oct_call(int mode);
int oct_ret(int mode);
int oct_hlt(int mode);

instruction oct_inst[] = {
  { "nop",  OPEC_NOP,  oct_nop },
  { "mov",  OPEC_MOV,  oct_mov },
  { "in",   OPEC_IN,   oct_in  },
  { "out",  OPEC_OUT,  oct_out },
  { "movi", OPEC_MOVI, oct_movi },
  { "addi", OPEC_ADDI, oct_addi },
  { "subi", OPEC_SUBI, oct_subi },
  { "muli", OPEC_MULI, oct_muli },
  { "divi", OPEC_DIVI, oct_divi },
  { "andi", OPEC_ANDI, oct_andi },
  { "ori",  OPEC_ORI,  oct_ori },
  { "testi",OPEC_TESTI,oct_testi },
  { "jmp",  OPEC_JMP,  oct_jmp },
  { "jz",   OPEC_JZ,   oct_jz },
  { "jnz",  OPEC_JNZ,  oct_jnz },
  { "jsr",  OPEC_JSR,  oct_jsr },
  { "push", OPEC_PUSH, oct_push },
  { "pop",  OPEC_POP,  oct_pop },
  { "call", OPEC_CALL, oct_call },
  { "ret",  OPEC_RET,  oct_ret },
  { "hlt",  OPEC_HLT,  oct_hlt },
  { "",           -1,       0 }
 };

// Speaker control parameters

#define PORT_PIT_CTRL	0x43	// 8253 Programmable Interval Timer control
#define PORT_PIT_CNT2	0x42	// PIT 2nd counter
#define PORT_8255B	0x61	// 8255 port B
#define PORT_WAIT	0x80	// 1 microsec wait / access
#define PIT_CMD		0xB6	// PIT control word
#define PIT_CLK2	1193180	// PIT clock frequency

void sound(void);

// Keyboard LED control

#define LED_NONE        0       // All LEDs are off
#define LED_SCROLL      1       // Scroll lock LED
#define LED_NUMBER      2       // Num lock LED
#define LED_CAPS        4       // Caps lock LED

int ledctrl(void);

// termios and Linux scan code definition

struct termios tm, ntm;		// "termios" structure holders

#define LINUXSCAN_ESC	1	// Linux scan code of "ESC"
#define LINUXSCAN_A	30	// Linux scan code of "A"
#define LINUXSCAN_K	37	// Linux scan code of "K"

// --- Program starts ---

int oct_init(void) {			// === "octopus" initializer ===
  int ret, i;

  // Open /dev/console

  fdcon = open("/dev/console", O_RDONLY);
  if (fdcon < 0) {
    fprintf(stderr, "/dev/console open failure!\n");
    return 1;
   }

  // Allow I/O access against PIT, port B, keyboard controler, and 0x80

  ret  = ioperm(PORT_PIT_CTRL, 1, 1);
  ret |= ioperm(PORT_PIT_CNT2, 1, 1);
  ret |= ioperm(PORT_8255B, 1, 1);
  ret |= ioperm(PORT_WAIT, 1, 1);
  if (ret) {
    fprintf(stderr, "I/O permission error!\n");
    return 1;
   }

  // Change keyboard mode to MEDIUMRAW (Linux scan code)

  if (enable_key) {
    ret = ioctl(fdcon, KDSKBMODE, K_MEDIUMRAW);
    if (ret < 0) {
      fprintf(stderr, "ioctl(KDSKBMODE) failure!\n");
      fprintf(stderr, "Sorry, you might activate me on X-window.\n");
      fprintf(stderr, "Please call me in a /dev/tty console!\n");
      return 1;
     }

  // Read current "termios" parameter

    ret  = tcgetattr(fdcon, &tm);
    ret |= tcgetattr(fdcon, &ntm);
    if (ret < 0) {
      fprintf(stderr, "tcgetattr() failure!\n");
      return 1;
     }

  // Setup realtime key scanning

    ntm.c_iflag &= ~(ISTRIP | INLCR | ICRNL | IGNCR | IXON | IXOFF);
    ntm.c_lflag &= ~(ECHO | ICANON | ISIG);
    ntm.c_cc[ VTIME ] = 0;	// We need no wait
    ntm.c_cc[ VMIN  ] = 0;	// Blank data is acceptable
    ret = tcsetattr(fdcon, TCSANOW, &ntm);
    if (ret < 0) {
      fprintf(stderr, "tcsetattr() failure!\n");
      return 1;
     }
   }

  // Initialize octopus

  pc = 0;				// Initialize pc = 0
  flag = 0;				// Reset flags

  for (i = 0; i < 256; i++)		// Clear port area with 255
    port[ i ] = 0xFF;
  port[ VPORT_SOUND ] = 0;		// Speaker is off
  port[ VPORT_FREQ_HIGH ] = 0x01;	// Set default frequency (440Hz)
  port[ VPORT_FREQ_LOW ] = 0xB8;
  port[ VPORT_KEY_LED ] = LED_NONE;	// Darkness at dawn
  port[ VPORT_KEY_AK ] = 0;		// All keys are depressed
  ret = ledctrl();			// Turn off all LEDs

  // Output initial memory contents

  message("  Memory = ");
  for (i = 0; i < fmt_text_bytes; i++) {
    if (i && (i % 16 == 0))
      message("           ");
    message("%02X ", text[i]);
    if (i % 16 == 15) message("\n");
   }
  if (i % 16) message("\n");
  message("----------------------------------------------------------\n");

  return ret;
 }

void oct_term(int dummy) {
  int i, ret, last;

  // Output memory and registers

  message("----------------------------------------------------------\n");
  message("            0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15\n");
  message("  Memory = ");
  for (i = 0; i < fmt_text_bytes; i++) {
    if (i && (i % 16 == 0))
      message("           ");
    message("%02X ", text[i]);
    if (i % 16 == 15) message("\n");
   }
  if (i % 16) message("\n");
  message("    Port = ");
  for (i = 0; i < fmt_port_bytes; i++) {
    if (i && (i % 16 == 0))
      message("           ");
    message("%02X ", port[i]);
    if (i % 16 == 15) message("\n");
   }
  if (i % 16) message("\n");
  message("Register = ");
  for (i = 0; i < 16; i++)
    message("%02X ", reg[i]);
  message("\n");
  message("    Flag = ");
  if (flag & ZFLAG)
    message("Z\n");
  else
    message("NZ\n");
  message("   Steps = %d\n", steps);

  // Restore the original "termios"

  if (enable_key) {
    ret = tcsetattr(fdcon, TCSANOW, &tm);
    if (ret < 0) {
      fprintf(stderr, "tcsetattr() failure!\n");
      ret = 1;
     }

  // Restore keyboard mode to "XLATE"

    ret = ioctl(fdcon, KDSKBMODE, K_XLATE);
    if (ret < 0) {
      fprintf(stderr, "ioctl(KDSKBMODE) failure!\n");
      ret = 1;
     }
   }

  // Sound off

  last = inb(PORT_8255B);		// Read last output value
  outb(last & ~3, PORT_8255B);		// Speaker off

  // Turn off all LEDs

  port[ VPORT_KEY_LED ] = 8;		// Turn off all LEDs and resume flags
  ret = ledctrl();
  close(fdcon);				// Close again
  exit(ret);				// Terminate process
 }

void disp_val(int val) {		// === value formatter ===
  char hexdec[] = "0123456789ABCDEF";
  char str[ 9 ];
  int w = val, r, idx = 8;

  if (fmt_base == 2) {			// Show number prefix
    imessage("0b");
   }
  else if (fmt_base == 16) {
    imessage("0x");
   }

  str[ idx ] = 0;			// Translate to string
  do {
    r = w % fmt_base;
    w /= fmt_base;
    str[ --idx ] = hexdec[ r ];
   }
  while (w > 0);

  if (fmt_zero_pad) {			// Do zero padding
    if (fmt_base == 2) {
      while (idx)
        str[ --idx ] = '0';
     }
    else if (fmt_base == 16) {
      if (val < 16)
        str[ --idx ] = '0';
     }
   }
  imessage("%s", &str[ idx ]);
 }

int oct_nop(int mod) {			// === "nop" instruction ===
  if (nop_wait) usleep(nop_wait);	// Do wait
  return 0;
 }

int oct_mov(int mod) {			// === "mov" instruction ===
  int src, dst;

  switch (mod) {
    case OPER_REG_REG:			// Register -> Register
      src = (text[ pc ] & 0xf0) >> 4;
      dst = text[ pc ] & 0x0f;
      imessage("r%d, r%d", src, dst);
      pc++;
      reg[ dst ] = reg[ src ];
      break;

    case OPER_REG_MEM:			// Register -> Memory
      imessage("r0, [ ");
      disp_val(text[ pc ]);
      imessage(" ]");
      pc++;
      text[ text[ pc - 1 ] ] = reg[ 0 ];
      break;

    case OPER_MEM_REG:			// Memory -> Register
      imessage("[ ");
      disp_val(text[ pc ]);
      imessage(" ], r0");
      pc++;
      reg[ 0 ] = text[ text[ pc - 1 ] ];
      break;

    case OPER_IND_REG:			// Reg indirect -> Register
      src = (text[ pc ] & 0xf0) >> 4;
      dst = text[ pc ] & 0x0f;
      imessage("[ r%d ], r%d", src, dst);
      pc++;
      reg[ dst ] = text[ reg[src] ];
      break;

    case OPER_REG_IND:			// Register -> Reg indirect
      src = (text[ pc ] & 0xf0) >> 4;
      dst = text[ pc ] & 0x0f;
      imessage("r%d, [ r%d ]", src, dst);
      pc++;
      text[ reg[dst] ] = reg[ src ];
      break;

    default:
      imessage("???\n");
      return OCT_ILLMOD;
   }
  return 0;
 }

int oct_in(int mod) {			// === "in" instruction ===
  int num;

  num = text[ pc++ ];			// Read port number from operand
  reg[ 0 ] = port[ num ];		// Transfer port data to R0
  imessage("[ ");
  disp_val(num);
  imessage(" ], r0");
  return 0;
 }

int oct_out(int mod) {			// === "out" instruction ===
  int num;

  num = text[ pc++ ];			// Read port number from operand
  port[ num ] = reg[ 0 ];		// Update port data with R0
  imessage("r0, [ ");
  disp_val(num);
  imessage(" ]");
  switch (num) {			// Which port?
    case VPORT_SOUND:			// Sound control
      sound();
      break;

    case VPORT_FREQ_HIGH:		// Sound frequency higher byte
      sound();
      break;

    case VPORT_FREQ_LOW:		// Sound frequency lower byte
      sound();
      break;

    case VPORT_KEY_LED:			// Keyboard LED control
      if (ledctrl()) return OCT_KEYERR;
      break;
   }
  return 0;
 }

int oct_movi(int dreg) {		// === "movi" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] = text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_addi(int dreg) {		// === "addi" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] += text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_subi(int dreg) {		// === "subi" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] -= text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_muli(int dreg) {		// === "muli" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] *= text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_divi(int dreg) {		// === "divi" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] /= text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_andi(int dreg) {		// === "andi" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] &= text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_ori(int dreg) {			// === "ori" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  reg[ dreg ] |= text[ pc - 1 ];
  if (reg[ dreg ] == 0)
    flag |= ZFLAG;
  else
    flag &= ~ZFLAG;
  return 0;
 }

int oct_testi(int dreg) {		// === "testi" instruction ===

  disp_val(text[ pc ]);
  imessage(", r%d", dreg);
  pc++;
  if (reg[ dreg ] & text[ pc - 1 ])
    flag &= ~ZFLAG;
  else
    flag |= ZFLAG;
  return 0;
 }

int oct_jmp(int mod) {			// === "jmp" instruction ===
  int dest;

  switch (mod) {
    case 0:
      dest = text[ pc ];		// Read jump address from operand
      disp_val(dest);
      break;

    case 1 ... 7:			// Register indirect jump
      dest = reg[ mod ];		// Read jump address from register
      imessage("r%d", mod);
      break;
   }
  pc = dest;				// Jump to target address
  return 0;
 }

int oct_jz(int mod) {			// === "jz" instruction ===
  int dest;

  switch (mod) {
    case 0:
      dest = text[ pc++ ];		// Read jump address from operand
      disp_val(dest);
      break;

    case 1 ... 7:			// Register indirect jump
      dest = reg[ mod ];		// Read jump address from register
      imessage("r%d", mod);
      break;
   }
  if (flag & ZFLAG)			// Zero flag is on ?
    pc = dest;				// Then, jump to target address
  return 0;
 }

int oct_jnz(int mod) {			// === "jnz" instruction ===
  int dest;

  switch (mod) {
    case 0:
      dest = text[ pc++ ];		// Read jump address from operand
      disp_val(dest);
      break;

    case 1 ... 7:			// Register indirect jump
      dest = reg[ mod ];		// Read jump address from register
      imessage("r%d", mod);
      break;
   }
  if ((flag & ZFLAG) == 0)		// Zero flag is off ?
    pc = dest;				// Then, jump to target address
  return 0;
 }

int oct_jsr(int mod) {			// === "jsr" instruction ===
  int dest;

  switch (mod) {
    case 0:
      dest = text[ pc++ ];		// Read jump address from operand
      disp_val(dest);
      break;

    case 1 ... 7:			// Register indirect jump
      dest = reg[ mod ];		// Read jump address from register
      imessage("r%d", mod);
      break;
   }
  rad = pc;				// Save return address
  pc = dest;				// Jump to target address
  return 0;
 }

int oct_push(int mod) {			// === "push" instruction ===
  int data;

  switch (mod) {
    case 0:
      data = text[ pc++ ];		// Read data in operand
      disp_val(data);
      break;

    case 1 ... 7:			// R1 - R7
      data = reg[ mod ];		// Read data from register
      imessage("r%d", mod);
      break;
   }
  text[ --sp ] = data;			// Save data in stack area
  return 0;
 }

int oct_pop(int mod) {			// === "pop" instruction ===
  int data;

  data = text[ sp++ ];			// Restore data from stack area
  switch (mod) {
    case 0:				// Illegal operand
      imessage("???\n");
      return OCT_ILLMOD;
      break;

    case 1 ... 7:			// R1 - R7
      imessage("r%d", mod);
      reg[ mod ] = data;		// Save data in register
      break;
   }
  return 0;
 }

int oct_call(int mod) {			// === "call" instruction ===
  int dest;

  switch (mod) {
    case 0:
      dest = text[ pc++ ];		// Read jump address from operand
      disp_val(dest);
      break;

    case 1 ... 7:			// Register indirect jump
      dest = reg[ mod ];		// Read jump address from register
      imessage("r%d", mod);
      break;
   }
  text[ --sp ] = pc;			// Save return address in stack area
  pc = dest;				// Jump to target address
  return 0;
 }

int oct_ret(int mod) {			// === "ret" instruction ===
  if (mod) {				// Illegal operand
      imessage("???\n");
      return OCT_ILLMOD;
   }
  pc = text[ sp++ ];			// Save destination add in stack area
  return 0;
 }

int oct_hlt(int mod) {			// === "hlt" instruction ===
  return OCT_HALTED;
 }

int oct_execute(void) {			// === "okto" decoder & activator ===
  int ins, op, i = 0, ret;

  ins = text[ pc ];			// Read instruction code
  op = (ins & 0xf8) >> 3;		// Get operation code number
  if (op == OPEC_BREAK)			// Break?
    return OCT_BREAK;			// Then do nothing
  pc++;					// Forward IP register
  if (pc == 0) return OCT_OVRRUN;	// Check over-run
  while (oct_inst[ i ].opecode != -1) {	// Search instruction table
    if (oct_inst[ i ].opecode == op) {	// Found!
      imessage("%02X: %s ", pc-1, oct_inst[ i ].name);
      ret = (*oct_inst[i].func)(ins&7);	// Execute the instruction
      imessage("\n");
      return ret;
     }
    i++;				// Next instruction
   }
  return OCT_ILLCOD;			// Illegal operation code
 }

// Sound port operation

void sound(void) {
  int freq, cnt, last;

  // Setup sound frequency

  freq = port[ VPORT_FREQ_HIGH ] * 256;	// Get MSB of sound frequency
  freq += port[ VPORT_FREQ_LOW ];	// Get LSB of sound frequency
  cnt = PIT_CLK2 / freq;		// Calculate timer 2 count
  outb(PIT_CMD, PORT_PIT_CTRL);		// Send a command to PIT
  inb(PORT_WAIT);			// Wait
  outb(cnt & 0xff, PORT_PIT_CNT2);	// Counter 2 LSB
  inb(PORT_WAIT);			// Wait
  outb((cnt&0xff00)>>8, PORT_PIT_CNT2);	// Counter 2 MSB

  // Speaker control

  last = inb(PORT_8255B);		// Read last output value
  if (port[ VPORT_SOUND ] & 1)		// Check speaker control bit
    outb(last | 3, PORT_8255B);		// Speaker on
  else
    outb(last & ~3, PORT_8255B);	// Speaker off
 }

// LED control routine

int ledctrl(void) {
  int stat, ret;

  stat = port[ VPORT_KEY_LED ];		// Read LED control value
  ret = ioctl(fdcon, KDSETLED, stat);
  if (ret < 0) {
    printf("ioctl(KDSETLED) failure!\n");
    return 1;
   }
  return 0;
 }

int main(int argc, char** argv) {	// === Everything starts here ===
  int idx = 1, i, ret, up, key, scan, bit;

  // Check option argument

  logfile = stderr;				// Default log is stderr
  while (idx < argc && argv[ idx ][ 0 ] == '-') {
    if (strcmp(argv[ idx ], "-key") == 0)
      enable_key = 1;				// Enable keyboard I/O
    else if (strcmp(argv[ idx ], "-time") == 0)
      max_exec_time = atoi(argv[ ++idx ]);	// Set time limit
    else if (strcmp(argv[ idx ], "-steps") == 0)
      max_exec_steps = atoi(argv[ ++idx ]);	// Set max steps
    else if (strcmp(argv[ idx ], "-wait") == 0)
      nop_wait = atoi(argv[ ++idx ]);		// Set NOP wait
    else if (strcmp(argv[ idx ], "-noinst") == 0)
      output_inst = 0;				// Turn off instruction
    else if (strcmp(argv[ idx ], "-text") == 0)
      fmt_text_bytes = atoi(argv[ ++idx ]);	// Text bytes to be displayed
    else if (strcmp(argv[ idx ], "-base") == 0)
      fmt_base = atoi(argv[ ++idx ]);		// Base value
    else if (strcmp(argv[ idx ], "-log") == 0) {
      logfile = fopen(argv[ ++idx ], "w");	// Open log file
      if (logfile == NULL)
        logfile = stderr;
     }
    idx++;
   }

  // Setup text area with specified values

  for (i = idx; i < argc; i++)
    text[ i - idx ] = atoi(argv[ i ]);
  text[ i - idx ] = 0xf8;		// Terminate with "hlt" instruction

  // Reset "octopus"

  if (oct_init())
    return 1;

  // Setup signal handlers

  signal(SIGHUP,  oct_term);
  signal(SIGINT,  oct_term);
  signal(SIGQUIT, oct_term);
  signal(SIGKILL, oct_term);
  signal(SIGALRM, oct_term);
  signal(SIGTERM, oct_term);

  // Execute instructions

  if (max_exec_time)
    alarm(max_exec_time);		// Set watchdog timer
  while (1) {
    ret = oct_execute();
    if (ret ==  OCT_BREAK) {
      message("### Software break at 0x%02X.\n", pc);
      break;
     }
    else if (ret ==  OCT_HALTED) {
      message("### System is halted.\n");
      break;
     }
    else if (ret == OCT_ILLCOD) {
      message("### Illegal operation code!\n");
      break;
     }
    else if (ret == OCT_ILLMOD) {
      message("### Illegal operation mode!\n");
      break;
     }
    else if (ret == OCT_OVRRUN) {
      message("### Segment over-run!\n");
      break;
     }
    else if (ret == OCT_KEYERR) {
      message("### Keyboard error!\n");
      break;
     };
    steps++;				// Increment step counter
    if (max_exec_steps &&		// Steps are limited
        steps >= max_exec_steps)	// and too many steps?
      raise(SIGINT);			// Then, abort

    // Handle keyboard event

    if (enable_key == 0)		// Keyboard I/O is disabled
      continue;
    ret = read(fdcon, &key, 1);		// Try to read a data
    if (ret == 0) 			// There is no data...
      continue;				// Then, scan again
    up = key & 0x80;			// Extract break bit
    scan = key & 0x7F;			// Extract linux scancode
    if (up && (scan == LINUXSCAN_ESC))	// ESC key was released
      break;				// Abort job
    else if (scan < LINUXSCAN_A || scan > LINUXSCAN_K)
      continue;				// Ignore it
    bit = 1 << (LINUXSCAN_K - scan);	// Calculate bit number
    if (up)				// This is break code
      port[ VPORT_KEY_AK ] &= ~bit;	// Turn off key bit
    else				// This is make code
      port[ VPORT_KEY_AK ] |= bit;	// Turn on key bit
   };
  raise(SIGINT);			// Send SIGINT signal to myself
  return 0;
 }
